#define F_CPU 16000000UL

#include <avr/io.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <stdlib.h>

#include "tft.h"
#include "font.h"

volatile unsigned int X_SIZE = 0;
volatile unsigned int Y_SIZE = 0;

void TFT9341_SendCommand(unsigned char cmd)
{  	CD_COMMAND;
  	DATA_PORT = cmd;
  	WR_STROBE;
}

void TFT9341_SendData(unsigned char data)
{  	CD_DATA;
 	DATA_PORT = data;
  	WR_STROBE;
}

void TFT9341_Write8(unsigned char data)
{  	DATA_PORT = data;
  	WR_STROBE;
}

void TFT9341_reset(void)
{	DATA_PORT 	= 0x00;
	DATA_DDR  	= 0xFF;
	COMDATA_DDR |= (1<<LCD_CD);
	CTRL1_DDR |= (1<<LCD_WR)|(1<<LCD_RD);
	CTRL2_DDR |= (1<<LCD_RESET);
  	WR_IDLE;
  	RD_IDLE;
  	RESET_ACTIVE;
  	_delay_ms(100);
  	RESET_IDLE;
  	TFT9341_SendCommand(0x01); // Software Reset
  	for (uint8_t i=0;i<3;i++) WR_STROBE;
}

void TFT9341_WriteRegister32(unsigned char r, unsigned long d)
{  	CD_COMMAND;
  	TFT9341_Write8(r);
  	CD_DATA;
  	_delay_us(1);
  	TFT9341_Write8(d>>24);
  	_delay_us(1);
  	TFT9341_Write8(d>>16);
  	_delay_us(1);
  	TFT9341_Write8(d>>8);
  	_delay_us(1);
  	TFT9341_Write8(d);
}

void TFT9341_SetAddrWindow(
	unsigned int x1,unsigned int y1,
	unsigned int x2,unsigned int y2)
{  	unsigned long t;
  	t = x1;
  	t<<=16;
  	t |= x2;
  	TFT9341_WriteRegister32(0x2A,t);//Column Addres Set
  	t = y1;
  	t<<=16;
  	t |= y2;
  	TFT9341_WriteRegister32(0x2B,t);//Page Addres Set
}

void TFT9341_SetRotation(unsigned char r)
{  	TFT9341_SendCommand(0x36);
	switch(r)
  	{  	case 0:
			TFT9341_SendData(0x48);
			X_SIZE = 240;
			Y_SIZE = 320;
    		break;
    	case 1:
			TFT9341_SendData(0x28);
			X_SIZE = 320;
			Y_SIZE = 240;
    		break;
    	case 2:
			TFT9341_SendData(0x88);
			X_SIZE = 240;
			Y_SIZE = 320;
    		break;
    	case 3:
			TFT9341_SendData(0xE8);
			X_SIZE = 320;
			Y_SIZE = 240;
    		break;
  	}
}

unsigned int Color565(unsigned char r, unsigned char g, unsigned char b)
{ 	unsigned int c = 0;
  	c = r >> 3;
  	c <<= 6;
  	c |= g >> 2;
  	c <<= 5;
  	c |= b >> 3;
  	return c;
}

void TFT9341_DrawPixel(int x, int y, unsigned int color)
{
//    
	if((x<0)||(y<0)||(x>=X_SIZE)||(y>=Y_SIZE)) return;
  	TFT9341_SetAddrWindow(x,y,X_SIZE-1,Y_SIZE-1);
  	CD_COMMAND;
  	TFT9341_Write8(0x2C);
  	CD_DATA;
  	TFT9341_Write8(color>>8);
	TFT9341_Write8(color);
}

void TFT9341_FillScreen(unsigned int color)
{  	TFT9341_SetAddrWindow(0,0,X_SIZE-1,Y_SIZE-1);
  	TFT9341_Flood(color,(long)X_SIZE*(long)Y_SIZE);
}


void TFT9341_FillRectangle(
	unsigned int color,					// 
	unsigned int x1, unsigned int y1,	//   
	unsigned int x2, unsigned int y2)
{	TFT9341_SetAddrWindow(x1, y1, x2, y2);
	TFT9341_Flood(color, (long)(x2-x1+1) * (long)(y2-y1+1));
}

void TFT9341_Flood(unsigned short color, unsigned long len)
{
  	unsigned short blocks;
  	unsigned char i,
  	hi = color, lo=(color>>8);
	CD_COMMAND;
 	TFT9341_Write8(0x2C);
	CD_DATA;
	TFT9341_Write8(hi);
	TFT9341_Write8(lo);
 
	TFT9341_Write8(lo);
	len--;

	blocks = (unsigned short)(len/64);//64 pixels/block
 
	if (hi == lo)
	{
		while(blocks--)
		{
  			i=16;
  			do
  			{
    			WR_STROBE;WR_STROBE;WR_STROBE;WR_STROBE;//2bytes/pixel
    			WR_STROBE;WR_STROBE;WR_STROBE;WR_STROBE;//x4 pixel
  			} while (--i);

  			for (i=(unsigned char)len&63;i--;)
  			{
    			WR_STROBE;
    			WR_STROBE;
  			}
		}
	}
	else
	{	while(blocks--)
		{	i=16;
  			do
  			{	TFT9341_Write8(hi);TFT9341_Write8(lo);TFT9341_Write8(hi);TFT9341_Write8(lo);
    			TFT9341_Write8(hi);TFT9341_Write8(lo);TFT9341_Write8(hi);TFT9341_Write8(lo);
  			} while (--i);
		}
	}

  	for (i = (unsigned char)len & 63; i--;)
  	{
    	TFT9341_Write8(hi);
    	TFT9341_Write8(lo);
  	}
}

void TFT9341_DrawLine(
	unsigned int color,
	unsigned int x1, unsigned int y1, 
	unsigned int x2, unsigned int y2)
{
	int steep = abs(y2 - y1) > abs(x2 - x1);
	if (steep)
	{	swap(x1, y1);
  		swap(x2, y2);
	}
	if (x1 > x2)
	{	swap(x1, x2);
  		swap(y1, y2);
	}
 
	int dx, dy;
 
	dx = x2 - x1;
	dy = abs(y2 - y1);
	int err = dx / 2;
	int ystep;

	if (y1 < y2)
	{
  		ystep = 1;
	} 
	else
	{
  		ystep = -1;
	}

  	for (; x1<=x2; x1++)
	{	if (steep)
		{
  			TFT9341_DrawPixel(y1, x1, color);
		} 
		else 
		{
  			TFT9341_DrawPixel(x1, y1, color);
		}

   		err -= dy;
     	if (err < 0)
		{
      		y1 += ystep;
      		err += dx;
    	}
	}
}


void TFT9341_Draw_Char(	//    1,2,3,4
	int x,			//  
	int y, 			//  
	unsigned int color, unsigned int backcolor, 
	unsigned char ch, 
	unsigned char size)
{
	unsigned char value = 0;
	unsigned char mask[8] =
	{	0b10000000,
		0b01000000,
		0b00100000,
		0b00010000,
		0b00001000,
		0b00000100,
		0b00000010,
		0b00000001
	};

 	if (ch>127) ch = ch-64;

 	switch(size)
  	{
    	int v,h;
    	int tv,th;
    	int tcolor;

		case 1:
  			for(h = 0; h < 5; h++)
//  			{	value = eeprom_read_byte(&(font5x8[ch-32][h]));
  			{	value = pgm_read_byte(&(font5x8[ch-32][h]));
//			value = pgm_read_byte(&font8x5[((unsigned char)ch) - 0x20][i]);

				for(v = 0; v < 8; v++)
				{	if ((value & mask[v]) != 0) tcolor = color;
					else tcolor = backcolor;
					TFT9341_DrawPixel(x+v,y+h,tcolor);
				}
  			}
  			break;
 
    	case 2:
  			for(h = 0; h < 5; h++)
//  			{	value = eeprom_read_byte(&(font5x8[ch-32][h]));
  			{	value = pgm_read_byte(&(font5x8[ch-32][h]));
				th = 2*h;
    			for(v = 0; v < 8; v++)			//   (x)
				{	tv = 2*v;
					if ((value & mask[v]) != 0) tcolor = color;
					else tcolor = backcolor;
					TFT9341_DrawPixel(x + tv,		y + th,		tcolor);
					TFT9341_DrawPixel(x + tv,		y + (th+1),	tcolor);
					TFT9341_DrawPixel(x + (tv+1),	y + th,		tcolor);
					TFT9341_DrawPixel(x + (tv+1),	y + (th+1),	tcolor);
  				}
			}
    		break;
  	}
}

void print_str(
		unsigned char x_pos, unsigned char y_pos,
		unsigned char font_size,
		unsigned int colour, unsigned int back_colour,
		char *str)
{	while (*str)
	{	TFT9341_Draw_Char(x_pos,y_pos,colour,back_colour,*str++,font_size);
		y_pos += (font_size * 5);
	}
}
